use super::*;

pub trait BinaryTreesAvl<'a>:
    BinaryTreesHeight<'a> + BinaryTreesTreeOf<'a> + BinaryTreesTryJoin<'a>
{
    fn assume_node(&self, tree: &Self::Tree) -> BTWrap<'a, Self, Self::Node> {
        match self.refer(tree) {
            Some(reference) => self.resolve(&reference),
            None => self.leaf_height_error(self.height(tree)),
        }
    }

    fn join_key_balanced_tree(
        &self,
        tl: Self::Tree,
        key: Self::Key,
        tr: Self::Tree,
    ) -> BTWrap<'a, Self, Self::Tree> {
        self.clone()
            .tree_bind(self.clone().join_key_balanced(tl, key, tr))
    }

    fn join_key_balanced(
        self,
        tl: Self::Tree,
        key: Self::Key,
        tr: Self::Tree,
    ) -> BTWrap<'a, Self, Self::Node> {
        let (hl, hr) = (self.height(&tl), self.height(&tr));
        match (hl.saturating_sub(hr), hr.saturating_sub(hl)) {
            (0, 0) | (0, 1) | (1, 0) => self.try_join(tl, key, tr),
            (0, _) => Self::bind(self.assume_node(&tr), move |nr| {
                let (trl, trr, kr) = self.split(&nr);
                let (rlh, rrh) = (self.height(&trl), self.height(&trr));
                if rlh > rrh {
                    Self::bind(self.assume_node(&trl), move |nrl| {
                        let (trll, trlr, krl) = self.split(&nrl);
                        Self::T::bind2(
                            self.join_key_balanced_tree(tl, key, trll),
                            self.join_key_balanced_tree(trlr, kr, trr),
                            |ti, to| self.join_key_balanced(ti, krl, to),
                        )
                    })
                } else {
                    Self::bind(self.join_key_balanced_tree(tl, key, trl), |t| {
                        self.join_key_balanced(t, kr, trr)
                    })
                }
            }),
            (_, 0) => Self::bind(self.assume_node(&tl), move |nl| {
                let (tll, tlr, kl) = self.split(&nl);
                let (hll, hlr) = (self.height(&tll), self.height(&tlr));
                if hll < hlr {
                    Self::bind(self.assume_node(&tlr), move |nlr| {
                        let (tlrl, tlrr, klr) = self.split(&nlr);
                        Self::T::bind2(
                            self.join_key_balanced_tree(tll, kl, tlrl),
                            self.join_key_balanced_tree(tlrr, key, tr),
                            |to, ti| self.join_key_balanced(to, klr, ti),
                        )
                    })
                } else {
                    Self::bind(self.join_key_balanced_tree(tlr, key, tr), |t| {
                        self.join_key_balanced(tll, kl, t)
                    })
                }
            }),
            (_, _) => unreachable!(),
        }
    }
}

impl<'a, BT: BinaryTreesHeight<'a> + BinaryTreesTreeOf<'a> + BinaryTreesTryJoin<'a>>
    BinaryTreesAvl<'a> for BT
{
}
